#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <fcntl.h>

#include "gen.h"
#include "sh.h"

Public char * * dissect();
Public char * smalloc();

Local bool escflg = True;	/*  interpret \ as "escape" character */

/*
 *	sh:  a Unix(tm) like command interpreter for MS-DOS.
 *
 *  by  Douglas Orr
 *      Textset Inc.
 *      Ann Arbor, Michigan
 *
 *  Copyright (c) 1985 Textset.   All rights reserved.
 *
 *  This program may be freely distributed, but not sold for profit.
 *
 */

main(argc,argv,envp)
int	argc;
char	* argv[];
char	* envp[];
	{

	/*  arguments processing */
	while( --argc )
	if( (*++argv)[0] == '-' )
		{
		switch( (*argv)[1] )
			{
			case 'e':
				escflg = !escflg;
				break;
			default:
				fprintf( stderr, "sh: invalid argument %s\n", 
					*argv );
				break;
			}
		}

	/*  real stuff  */
	init();
	sh();

	}


/*  Environment Stuff  */

/*
 *   for now, just use the uSoft environment stuff.  This could probably
 *  be improved
 */
char *
l_getenv( name )
char * name;
	{
	return( getenv(name) );
	}

l_setenv( name, val )
char * name;
char * val;
	{

	char * str;

	str = smalloc( strlen(name) + strlen(val) + 2 );

	strcpy( str, name );
	strcat( str, "=" );
	strcat( str, val );

	return( putenv( str ) );

	}

l_printenv()
	{

	char * * ep;

	putchar( '\n' );
	for( ep=environ; *ep; ep++ )
		printf( "%s\n", *ep );
	putchar( '\n' );

	return( 0 );

	}


init()
	{

	char * path;

	/*  ignore interrupts */
	signal( SIGINT, SIG_IGN );

	/*   default values  */
	if( (path = l_getenv( "PATH" )) == NULL )
		{
		l_setenv( "PATH", "\;\bin" );
		path = l_getenv( "PATH" );
		}

	setbuf( stdin, NULL );
	setbuf( stdout, NULL );
	setbuf( stderr, NULL );

	rehash( path );

	/*   read in the profile  */
	if( access( "/profile", F_ROK ) == 0 )
		source( "/profile" );

	/*  set important variables  */
	if( l_getenv( "PROMPT" ) == NULL )
		l_setenv( "PROMPT", "% " );
	if( l_getenv( "PS1" ) == NULL )
		l_setenv( "PS1", "> " );

	}


/*
 *	iiiiit's SHOWTIME
 */
sh()
	{
	char buf[256];
	bool intr;

	while( !feof(stdin) )
		{

		/*  prompt  */
		if( Interactive() )
			fprintf( stderr, "%s", l_getenv("PROMPT") );

		/*  input  */
		if( getbuf( buf, sizeof buf, stdin ) )
			{
			/*  execute  */
			(void)sys( buf, sizeof buf );
			}

		}

	}

/*
 *  Do initial input processing.  Turn the high bit on to
 * prevent escaped characters and special characters within single
 * quotes from being interpreted
 */

#define	ChNorm (0)
#define	ChDquote (1)
#define	ChSquote (2)

#define	Escape(x) (x|0x80)

getbuf( buf, bufsize, fd )
char * buf;
int bufsize;
FILE * fd;
	{
	char * ptr = buf;
	int ch_state;
	Reg int ch;

	ch_state = ChNorm;
	while( (ch = getc(fd)) != EOF )
		{
		if( (ch == '\\') && escflg )
			ch = Escape(getc(fd));
		else
		if( ch == '\'' )
			{
			if( ch_state == ChSquote )
				ch_state = ChNorm;
			else
			if( ch_state == ChNorm )
				ch_state = ChSquote;
			else
			/*  state == dquote */
				ch = Escape(ch);
			}
		else
		if( ch == '"' )
			{
			if( ch_state == ChDquote )
				ch_state = ChNorm;
			else
			if( ch_state == ChNorm )
				ch_state = ChDquote;
			else
			/*  state == squote */
				ch = Escape(ch);
			}

		/* ?!  include other wildcards here, as added  */
		if( ((ch_state == ChSquote) &&  ((ch == '$') || (ch == '*')))
		||  ((ch_state == ChDquote) &&  (ch == '*')) )
			ch = Escape(ch);

		*ptr++ = ch;
		if( ptr-buf >= bufsize-1 ) 
			break;
		/*
		 *  !! This isn't exactly how Unix does quote processing ...
		 *   my current opinion is that it's not worth the trouble
		 */
		if( ch == '\n' )
			{
			if( ch_state == ChNorm )
				break;
			else
			if( Interactive() )
				fprintf( stderr, "%s ", l_getenv( "PS1" ) );
			}

		}

	*ptr = '\0';

	/*  eof */
	if( ptr == buf )
		return( False );

	/*  screw up  */
	if( ch_state != ChNorm )
		{
		fprintf( stderr, "error: mismatched quotes\n" );
		return( False );
		}

	return( True );

	}


/*
 *	Keep a stack of fds.  Push our current I/O environment before
 *  executing commands so that we have someplace to go back to after
 *  I/O redirections
 */
#define FMax (20)
Local int fdstk[FMax];
Local int ftop = 0;

push_fd(fd)
int fd;
	{
	if( ftop >= FMax )
		return( -1 );
	else
		fdstk[ftop++] = fd;
	}

pop_fd()
	{
	if( ftop == 0 )
		return( -1 );
	else
		return( fdstk[--ftop] );
	}


save_fds()
	{
	push_fd( dup(0) );
	push_fd( dup(1) );
	push_fd( dup(2) );
	}

restore_fds()
	{
	int fd;

	fflush(stderr);
	fflush(stdout);

	dup2( (fd = pop_fd()), 2 );
	close(fd);
	dup2( (fd = pop_fd()), 1 );
	close(fd);
	dup2( (fd = pop_fd()), 0 );
	close(fd);

	setbuf( stdin, NULL );
	clearerr(stdin);
	clearerr(stdout);
	clearerr(stderr);

	}


/*
 *   Use as source input the indicated file
 */
source( file )
char * file;
	{
	int fd;


	if( (fd = open( file, 0 )) >= 0 )
		{

		/*   save current input  */
		if( push_fd( dup(0) ) == -1 )
			{
			fprintf( stderr, "source files nested too deeply\n" );
			return;
			}
		dup2( fd, 0 );
		close(fd);

		/*  loop on new input  */
		sh();

		/*  restore old input  */
		dup2( (fd = pop_fd()), 0 );
		clearerr(stdin);
		close(fd);

		}
	else
		perror( file );

	}


/*
 *	execute the given command
 */
sys( pgm, pgmlen )
char * pgm;
int pgmlen;
	{

	int rc;
	char * * argv;

	/*  set up I/O environment  */
	save_fds();

	/*  slice into an argv  */
	if( (argv = dissect( pgm, pgmlen )) )
		{
		/*  doit  */
		rc = cmd( argv, environ );

		/*  !! Assertion:  all args are malloc'd */
		while( *argv )
			free( *argv++ );
		}

	/*  restore  */
	restore_fds();

	return( rc );

	}
